/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.io;

import jalview.datamodel.DBRefEntry;
import jalview.datamodel.SequenceFeature;
import jalview.datamodel.SequenceI;
import jalview.io.gff.GffConstants;
import jalview.util.UrlLink;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * generate HTML reports for a sequence
 * 
 * @author jimp
 */
public class SequenceAnnotationReport
{
  final String linkImageURL;

  public SequenceAnnotationReport(String linkImageURL)
  {
    this.linkImageURL = linkImageURL;
  }

  /**
   * Append text for the list of features to the tooltip
   * 
   * @param tooltipText2
   * @param rpos
   * @param features
   * @param minmax
   */
  public void appendFeatures(final StringBuffer tooltipText2, int rpos,
          List<SequenceFeature> features, Map<String, float[][]> minmax)
  {
    if (features != null)
    {
      for (SequenceFeature feature : features)
      {
        appendFeature(tooltipText2, rpos, minmax, feature);
      }
    }
  }

  /**
   * Appends text for one sequence feature to the string buffer
   * 
   * @param sb
   * @param rpos
   * @param minmax
   *          {{min, max}, {min, max}} positional and non-positional feature
   *          scores for this type
   * @param feature
   */
  void appendFeature(final StringBuffer sb, int rpos,
          Map<String, float[][]> minmax, SequenceFeature feature)
  {
    if (feature.isContactFeature())
    {
      if (feature.getBegin() == rpos || feature.getEnd() == rpos)
      {
        if (sb.length() > 6)
        {
          sb.append("<br>");
        }
        sb.append(feature.getType()).append(" ").append(feature.getBegin())
                .append(":")
                .append(feature.getEnd());
      }
    }
    else
    {
      if (sb.length() > 6)
      {
        sb.append("<br>");
      }
      // TODO: remove this hack to display link only features
      boolean linkOnly = feature.getValue("linkonly") != null;
      if (!linkOnly)
      {
        sb.append(feature.getType()).append(" ");
        if (rpos != 0)
        {
          // we are marking a positional feature
          sb.append(feature.begin);
        }
        if (feature.begin != feature.end)
        {
          sb.append(" " + feature.end);
        }

        if (feature.getDescription() != null
                && !feature.description.equals(feature.getType()))
        {
          String tmpString = feature.getDescription();
          String tmp2up = tmpString.toUpperCase();
          final int startTag = tmp2up.indexOf("<HTML>");
          if (startTag > -1)
          {
            tmpString = tmpString.substring(startTag + 6);
            tmp2up = tmp2up.substring(startTag + 6);
          }
          // TODO strips off </body> but not <body> - is that intended?
          int endTag = tmp2up.indexOf("</BODY>");
          if (endTag > -1)
          {
            tmpString = tmpString.substring(0, endTag);
            tmp2up = tmp2up.substring(0, endTag);
          }
          endTag = tmp2up.indexOf("</HTML>");
          if (endTag > -1)
          {
            tmpString = tmpString.substring(0, endTag);
          }

          if (startTag > -1)
          {
            sb.append("; ").append(tmpString);
          }
          else
          {
            if (tmpString.indexOf("<") > -1 || tmpString.indexOf(">") > -1)
            {
              // The description does not specify html is to
              // be used, so we must remove < > symbols
              tmpString = tmpString.replaceAll("<", "&lt;");
              tmpString = tmpString.replaceAll(">", "&gt;");
              sb.append("; ").append(tmpString);
            }
            else
            {
              sb.append("; ").append(tmpString);
            }
          }
        }

        /*
         * score should be shown if there is one, and min != max
         * for this feature type (e.g. not all 0)
         */
        if (!Float.isNaN(feature.getScore()))
        {
          float[][] rng = (minmax == null) ? null : minmax.get(feature
                  .getType());
          if (rng != null && rng[0] != null && rng[0][0] != rng[0][1])
          {
            sb.append(" Score=").append(String.valueOf(feature.getScore()));
          }
        }
        String status = (String) feature.getValue("status");
        if (status != null && status.length() > 0)
        {
          sb.append("; (").append(status).append(")");
        }
        String clinSig = (String) feature
                .getValue(GffConstants.CLINICAL_SIGNIFICANCE);
        if (clinSig != null)
        {
          sb.append("; ").append(clinSig);
        }
      }
    }
    appendLinks(sb, feature);
  }

  /**
   * Format and appends any hyperlinks for the sequence feature to the string
   * buffer
   * 
   * @param sb
   * @param feature
   */
  void appendLinks(final StringBuffer sb, SequenceFeature feature)
  {
    if (feature.links != null)
    {
      if (linkImageURL != null)
      {
        sb.append(" <img src=\"" + linkImageURL + "\">");
      }
      else
      {
        for (String urlstring : feature.links)
        {
          try
          {
            for (List<String> urllink : createLinksFrom(null, urlstring))
            {
              sb.append("<br/> <a href=\""
                      + urllink.get(3)
                      + "\" target=\""
                      + urllink.get(0)
                      + "\">"
                      + (urllink.get(0).toLowerCase()
                              .equals(urllink.get(1).toLowerCase()) ? urllink
                              .get(0) : (urllink.get(0) + ":" + urllink
                              .get(1)))
                      + "</a></br>");
            }
          } catch (Exception x)
          {
            System.err.println("problem when creating links from "
                    + urlstring);
            x.printStackTrace();
          }
        }
      }

    }
  }

  /**
   * 
   * @param seq
   * @param link
   * @return Collection< List<String> > { List<String> { link target, link
   *         label, dynamic component inserted (if any), url }}
   */
  Collection<List<String>> createLinksFrom(SequenceI seq, String link)
  {
    Map<String, List<String>> urlSets = new LinkedHashMap<String, List<String>>();
    UrlLink urlLink = new UrlLink(link);
    if (!urlLink.isValid())
    {
      System.err.println(urlLink.getInvalidMessage());
      return null;
    }

    urlLink.createLinksFromSeq(seq, urlSets);

    return urlSets.values();
  }

  public void createSequenceAnnotationReport(final StringBuffer tip,
          SequenceI sequence, boolean showDbRefs, boolean showNpFeats,
          Map<String, float[][]> minmax)
  {
    createSequenceAnnotationReport(tip, sequence, showDbRefs, showNpFeats,
            true, minmax);
  }

  public void createSequenceAnnotationReport(final StringBuffer tip,
          SequenceI sequence, boolean showDbRefs, boolean showNpFeats,
          boolean tableWrap, Map<String, float[][]> minmax)
  {
    String tmp;
    tip.append("<i>");

    int maxWidth = 0;
    if (sequence.getDescription() != null)
    {
      tmp = sequence.getDescription();
      tip.append("<br>" + tmp);
      maxWidth = Math.max(maxWidth, tmp.length());
    }
    SequenceI ds = sequence;
    while (ds.getDatasetSequence() != null)
    {
      ds = ds.getDatasetSequence();
    }
    DBRefEntry[] dbrefs = ds.getDBRefs();
    if (showDbRefs && dbrefs != null)
    {
      for (int i = 0; i < dbrefs.length; i++)
      {
        tip.append("<br>");
        tmp = dbrefs[i].getSource() + " " + dbrefs[i].getAccessionId();
        tip.append(tmp);
        maxWidth = Math.max(maxWidth, tmp.length());
      }
    }

    // ADD NON POSITIONAL SEQUENCE INFO
    SequenceFeature[] features = sequence.getSequenceFeatures();
    if (showNpFeats && features != null)
    {
      for (int i = 0; i < features.length; i++)
      {
        if (features[i].begin == 0 && features[i].end == 0)
        {
          int sz = -tip.length();
          List<SequenceFeature> tfeat = new ArrayList<SequenceFeature>();
          tfeat.add(features[i]);
          appendFeatures(tip, 0, tfeat, minmax);
          sz += tip.length();
          maxWidth = Math.max(maxWidth, sz);
        }
      }
    }

    if (tableWrap && maxWidth > 60)
    {
      tip.insert(0, "<table width=350 border=0><tr><td><i>");
      tip.append("</i></td></tr></table>");
    }

  }
}
